"""\
Automatically convert Apress Word Format to reStructuredText (via .odt)

This is a tool for managing the conversion processes for converting from a .odt
file obtained from an Apress Word document into reStructuredText.

Use like this:

::

    $ unoconv -f odt 9349ch*_backout.doc
    $ python process2.py *.odt

Features:

* Converts every type of formatting used in the book "The Definitive Guide To
  Pylons" (apart from the items discussed in the Limitiaitons). Includes:
  * Paragraphs
  * Bulleted Lists
  * Numbered Lists
  * Single Code Examples
  * Multiline Code Examples
  * Figures
  * Tables
  * Notes, Tips, Cautions
  * Emphasis
  * Literals
  * Bold

Limitations:

* No support for bold or italic code examples, this isn't something 
  reStructuredText supports so no warnings are raised.
* Some admonitions such as Note, Tip or Caution are generated incorrectly
  because the Apress Word document doesn't contain enough semantic markup 
  to be able to automatically determine the correct thing to do
* ~CCC characters behave strangely
* Some multiline code examples don't have the correct markup for their last
  line so no newline is added, generating a warning when the reStructuredText
  is parsed (HTML is still generated correctly though)
* Sphinx only supports index entries at the paragraph level so 
* Definition lists in the Apress template do not contain semantic markup so
  we have to guess when they are really lists by checking for a : character in
  the first 100 characters. This breaks when a URL stating http:// is also
  in the first 100 characters.
"""

import os
import os.path
import sys
import zipfile
import xml.dom.minidom
import re

import table

#
# URL Matching
#

urlfinders = [
    re.compile("([0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}|(((news|telnet|nttp|file|http|ftp|https)://)|(www|ftp)[-A-Za-z0-9]*\\.)[-A-Za-z0-9\\.]+)(:[0-9]*)?/[-A-Za-z0-9_\\$\\.\\+\\!\\*\\(\\),;:@&=\\?/~\\#\\%]*[^]'\\.}>\\),\\\"]"),
    re.compile("([0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}|(((news|telnet|nttp|file|http|ftp|https)://)|(www|ftp)[-A-Za-z0-9]*\\.)[-A-Za-z0-9\\.]+)(:[0-9]*)?"),
    #re.compile("(~/|/|\\./)([-A-Za-z0-9_\\$\\.\\+\\!\\*\\(\\),;:@&=\\?/~\\#\\%]|\\\\)+"),
    #re.compile("'\\<((mailto:)|)[-A-Za-z0-9\\.]+@[-A-Za-z0-9\\.]+"),
]

def link(content):
    for regex in urlfinders:
        match = regex.search(content)
        if match is not None:
            url = content[match.start(): match.end()]
            return content[:match.start()]+'`%s <%s>`_'%(url, url)+content[match.end():]
    return content

#
# Main DOM Handler
#

class ChapterConverter(object):
    output_level = 0
    context = ['normal']
    cur_output = ''

    def output(self, text=''):
        for line in text.split('\n'):
            self.cur_output += ("\t"*self.output_level)+line+'\n'
   
    def get_text(self, nodelist):
        if not isinstance(nodelist, list):
            nodelist = nodelist.childNodes
        rc = ""
        for node in nodelist:
            if node.nodeType == node.TEXT_NODE:
                rc = rc + node.data
        return rc

    def handle_chapter(self, chapter):
        chapter_title = "%s: %s" % (
            self.get_text(chapter.childNodes[0]),
            self.get_text(chapter.childNodes[1]),
        )
        self.output(chapter_title)
        self.output('+'*len(chapter_title))
        self.output()
        block = self.handle_block(chapter.childNodes[2:])
        self.output(block)
        return self.cur_output

    def handle_block(self, nodes):
        collected_nodes = []
        result = ''
        collect_nodes = False
        mode = ''
        for child in nodes:
            if not collect_nodes:
                if child.nodeName == 'p' and child.attributes['style-name'].value in ['Body_20_Text_20_First', 'Text_20_body', 'Body_20_Text_20_First_20_copy', 'Body_20_Text_20_Cont']:
                    paragraph = self.handle_paragraph(child)
                    result += paragraph+'\n\n'
                elif child.nodeName == 'p' and child.attributes['style-name'].value in ['Note_2f_Tip_2f_Caution']:
                    # Note/Tip/Caution boxes are expected to be wrapped in Begin Note / End Note production tags
                    # or to be a single paragraph consisiting of:
                    # [<DOM Text node "Note">, <DOM Element: tab at 0x114c050>, <DOM Text node "If you don...">]
                    box_elems = child.childNodes
                    valid_box = True
                    if not len(box_elems) >= 3 or \
                       box_elems[0].nodeType != child.TEXT_NODE or \
                       box_elems[1].nodeName != 'tab':
                        sys.stderr.write(
                            "ERROR: Found a 'Note_2f_Tip_2f_Caution' "
                            'element without the expected structure. It is '
                            'likely you forgot the Begin Note/End Node '
                            'production tags on a multi-paragarph note, '
                            'tip or caution. Treating it as a paragraph '
                            'containg the following nodes: %r\n'%(
                                self.get_text(box_elems).encode('utf-8')
                            )
                        )
                        valid_box = False
                    else: 
                        box_type = box_elems[0].data.lower()
                        if not box_type in ['caution', 'tip', 'note']:
                            sys.stderr.write('CRITICAL: Unknown box type %s\n'%box_type.encode('utf-8'))
                            box_type = 'note'
                        # Now that we have the box type, let's remove those nodes, and turn the element into a paragraph for normal processing
                        child.removeChild(box_elems[0])
                        child.removeChild(box_elems[0])
                    child.attributes['style-name'].value = 'Text_20_body'
                    paragraph = self.handle_paragraph(child)
                    # Some single notes have a control character to mean there should actually be another paragraph
                    paragraph = paragraph.replace('~MS ', '\n\n')
                    if valid_box:
                        result += '.. %s ::\n\n'%box_type
                        for line in paragraph.split('\n'):
                            result += '\t'+line+'\n'
                        result += '\n'
                    else:
                        result += paragraph+'\n\n'
                elif child.nodeName == 'p' and child.attributes['style-name'].value in ['Code_20_Single']:
                    code = self.handle_code_single(child)
                    result += code+'\n\n'
                elif child.nodeName == 'p' and child.attributes['style-name'].value in ['Code_20_First', 'P4', 'P1']:
                    code = self.handle_code_single(child)
                    result += code+'\n'
                # We don't support bold code.
                elif child.nodeName == 'p' and child.attributes['style-name'].value in ['Code', 'P5', 'Code_20_Bold']:
                    code = self.handle_code_line(child)
                    result += code+'\n'
                elif child.nodeName == 'p' and child.attributes['style-name'].value in ['Code_20_Last', 'Code_20_Last_20_Bold']:
                    code = self.handle_code_line(child)
                    result += code+'\n\n'
                elif child.nodeName == 'p' and child.attributes['style-name'].value == 'Quote':
                    result += '\n\n\t'+self.get_text(child)+'\n\n'
                elif child.nodeName == 'p' and child.attributes['style-name'].value == 'P1':
                    # PI elements are used for italics in Chapter 1, treat them as code
                    code = self.handle_code_line(child)
                    result += code+'\n'
                elif child.nodeName == 'h':
                    if child.attributes['style-name'].value  == 'Heading_20_1':
                        heading = self.handle_heading(child, '=')
                    elif child.attributes['style-name'].value  == 'Heading_20_2':
                        heading = self.handle_heading(child, '-')
                    elif child.attributes['style-name'].value  == 'Heading_20_3':
                        heading = self.handle_heading(child, '~')
                    else: 
                        self.unexpected_child('chapter', child)
                    result += heading+'\n\n'
                elif child.nodeName == 'p' and child.attributes['style-name'].value == "FM_20_Head":
                    heading = self.handle_heading(child, '=')
                    result += heading+'\n\n'
                elif child.nodeName == 'p' and child.attributes['style-name'].value == "Bullet_20_Sub_20_List":
                    sys.stderr.write('CRITICAL: Can\'t handle badly labelled sub-lists (Bullet_20_Sub_20_List). Content ignored\n')
                elif child.nodeName == 'p' and child.attributes['style-name'].value == "Bullet_20_First":
                    collect_nodes = True
                    mode = 'bullet_list'
                    collected_nodes.append(child)
                elif child.nodeName == 'p' and child.attributes['style-name'].value == "Unnumbered_20_First":
                    collect_nodes = True
                    mode = 'item_list'
                    collected_nodes.append(child)
                elif child.nodeName == 'p' and child.attributes['style-name'].value == "Unnumbered_20_Last":
                    # Its possible we have a one-item definition list.
                    child.attributes['style-name'].value = 'Text_20_body'
                    result += self.handle_item_list_entry([child])
                elif child.nodeName == 'p' and child.attributes['style-name'].value == "Table_20_Caption":
                    collect_nodes = True
                    mode = 'table'
                    collected_nodes.append(child)
                elif child.nodeName == 'p' and child.attributes['style-name'].value == "Num_20_List_20_First":
                    collect_nodes = True
                    mode = 'enumerated_list'
                    collected_nodes.append(child)
                elif child.nodeName == 'p' and child.attributes['style-name'].value == "Production":
                    if self.get_text(child) == 'Begin Note':
                        collect_nodes = True
                        mode = 'note'
                        collected_nodes.append(child)
                    elif self.get_text(child).startswith('Insert '):
                        collect_nodes = True
                        mode = 'figure'
                        collected_nodes.append(child)
                    else:
                        self.unexpected_child('chapter', child)
                else:
                    self.unexpected_child('chapter', child)
            else:
                if mode == 'figure':
                    if child.nodeName == 'p' and child.attributes['style-name'].value == 'Figure_20_Caption':
                        collected_nodes.append(child)
                        # The two nodes contain the URL and the caption 
                        url = self.get_text(collected_nodes[0]).replace('Insert ', '').replace('r', '').replace('tiff', 'png').replace('tif', 'png')
                        collected_nodes[1].attributes['style-name'].value = 'Text_20_body'
                        caption = self.handle_paragraph(collected_nodes[1])
                        result += '.. figure :: %s\n'%url
                        result += '   :target: _images/%s\n'%url
                        result += '   :scale: 70\n'
                        result += '   :alt: %s\n\n'%caption.replace('\n', '')
                        result += '   %s\n\n'%caption
                        # Reset the collection variables
                        collect_nodes = False
                        collected_nodes = []
                        mode = ''
                    else:
                        url = self.get_text(collected_nodes[0]).replace('Insert ', '').replace('r', '').replace('tiff', 'png').replace('tif', 'png').replace('.eps', '.png')
                        result += '.. figure :: %s\n'%url
                        sys.stderr.write('CRITICAL: Missed a node because found an image instead of a figure\n')
                        # Reset the collection variables
                        collect_nodes = False
                        collected_nodes = []
                        mode = ''
                elif mode == 'table' and child.nodeName == 'p' and child.attributes['style-name'].value == 'Table_20_Text_20_Last':
                    collected_nodes.append(child)
                    result += self.handle_table(collected_nodes)
                    # Reset the collection variables
                    collect_nodes = False
                    collected_nodes = []
                    mode = ''
                elif child.attributes['style-name'].value == "Production":
                    if mode == 'note' and child.nodeName == 'p' and self.get_text(child) == 'End Note':
                        collected_nodes.append(child)
                        # Some of the blocks within here will be called 'Note_2f_Tip_2f_Caution' when they should just be treated as paragraphs:
                        block = []
                        result += self.handle_block([collected_nodes[1]])
                        for box_node in collected_nodes[2:-1]:
                            if box_node.attributes['style-name'].value == 'Note_2f_Tip_2f_Caution':
                                box_node.attributes['style-name'].value = 'Text_20_body'
                            block.append(box_node)
                        content = self.handle_block(block)
                        #result += '.. note ::\n\n'
                        for line in content.split('\n'):
                            result += '\t'+line+'\n'
                        # Reset the collection variables
                        collect_nodes = False
                        collected_nodes = []
                        mode = ''
                    else:
                        collected_nodes.append(child)
                elif mode == 'enumerated_list' and child.nodeName == 'p' and child.attributes['style-name'].value == "Num_20_List_20_Last":
                    collected_nodes.append(child)
                    # Need to split the nodes into groups of each one with a starting bullet.
                    bullets = [] 
                    current_bullet = []
                    # Find each paragrpah, check it starts with <tab>1.<tab>. 
                    for item in collected_nodes:
                        if item.nodeName == 'p' and item.attributes['style-name'].value in ['Num_20_List', 'Num_20_List_20_Last', 'Num_20_List_20_First']:
                            bullets.append(current_bullet)
                            item.removeChild(item.childNodes[0])
                            item.removeChild(item.childNodes[0])
                            item.removeChild(item.childNodes[0])
                            item.attributes['style-name'].value = 'Text_20_body'
                            current_bullet = [item]
                        else:
                            current_bullet.append(item)
                    # Indent the first line of each set of results by a tab a * and a tab, indent everything else with two tabs.
                    bullets.append(current_bullet)
                    counter = 1
                    for bullet_list in bullets[1:]:
                        bullet_parts = self.handle_block(bullet_list).split('\n')
                        result += '\t'+str(counter)+'.\t'+bullet_parts[0]+'\n\n'
                        for line in bullet_parts[1:]:
                            if line.strip():
                                result += '\t\t'+line+'\n\n'
                        result = result[:-1]
                        counter += 1
                    result += '\n'
                    # Reset the collection variables
                    collect_nodes = False
                    collected_nodes = []
                    mode = ''
                elif mode == 'item_list' and child.nodeName == 'p' and child.attributes['style-name'].value == "Unnumbered_20_Last":
                    collected_nodes.append(child)
                    # Need to split the nodes into groups of each one with a starting bullet.
                    bullets = [] 
                    current_bullet = []
                    # Find each paragrpah, check it starts with <tab>*<tab>. 
                    for item in collected_nodes:
                        if item.nodeName == 'p' and item.attributes['style-name'].value in ['Unnumbered_20_First', 'Unnumbered_20_Last']:
                            bullets.append(current_bullet)
                            item.attributes['style-name'].value = 'Text_20_body'
                            current_bullet = [item]
                        elif item.nodeName == 'p' and item.attributes['style-name'].value in ['Unnumbered']:
                            if item.childNodes[0].nodeType == child.TEXT_NODE and ':' not in item.childNodes[0].data[:100]:
                                # Treat as a paragraph
                                item.attributes['style-name'].value = 'Text_20_body'
                                current_bullet.append(item)
                            # It is possible this is a paragraph if there isn't anything after the ``:``
                            elif item.childNodes[0].nodeType == child.TEXT_NODE and not item.childNodes[0].data.split(':')[1].replace('\n','').strip():
                                # Treat as a paragraph
                                item.attributes['style-name'].value = 'Text_20_body'
                                current_bullet.append(item)
                            else:
                                bullets.append(current_bullet)
                                item.attributes['style-name'].value = 'Text_20_body'
                                current_bullet = [item]
                        else:
                            current_bullet.append(item)
                    # Use the paragraph tool on the first 
                    bullets.append(current_bullet)
                    for bullet_list in bullets[1:]:
                        result += self.handle_item_list_entry(bullet_list)
                    result += '\n'
                    # Reset the collection variables
                    collect_nodes = False
                    collected_nodes = []
                    mode = ''
                elif mode == 'bullet_list' and child.nodeName == 'p' and child.attributes['style-name'].value == "Bullet_20_Last":
                    collected_nodes.append(child)
                    # Need to split the nodes into groups of each one with a starting bullet.
                    bullets = [] 
                    current_bullet = []
                    # Find each paragrpah, check it starts with <tab>*<tab>. 
                    for item in collected_nodes:
                        if item.nodeName == 'p' and item.attributes['style-name'].value in ['Bullet_20_First', 'Bullet_20_Last', 'Bullet']:
                            bullets.append(current_bullet)
                            item.removeChild(item.childNodes[0])
                            item.removeChild(item.childNodes[0])
                            item.removeChild(item.childNodes[0])
                            item.attributes['style-name'].value = 'Text_20_body'
                            current_bullet = [item]
                        else:
                            current_bullet.append(item)
                    # Indent the first line of each set of results by a tab a * and a tab, indent everything else with two tabs.
                    bullets.append(current_bullet)
                    for bullet_list in bullets[1:]:
                        bullet_parts = self.handle_block(bullet_list).split('\n')
                        result += '\t*\t'+bullet_parts[0]+'\n\n'
                        for line in bullet_parts[1:]:
                            if line.strip():
                                result += '\t\t'+line+'\n\n'
                        result = result[:-1]                        
                    result += '\n'
                    # Reset the collection variables
                    collect_nodes = False
                    collected_nodes = []
                    mode = ''
                else:
                    collected_nodes.append(child)
        if collect_nodes is True:
            sys.stderr.write('CRITICAL: Failed to find expected end tag so couldn\'t process the rest of the document: %r\n'%collected_nodes[0].childNodes)
        return result

    def handle_item_list_entry(self, nodes):
        result = ''
        bullet_parts = self.handle_block(nodes).split('\n\n')
        # What we need to do here is split the content on the : character and parse each as paragraphs
        index = None
        if bullet_parts[0].startswith('.. index'):
            # There is an index entry to worry about.
            index = bullet_parts[0]
            bullet_parts = bullet_parts[1:]
        parts = bullet_parts[0].split(':')
        key = parts[0]
        value = ':'.join(parts[1:])
        if value and value[0] == ' ':
            value = value[1:]+'\n'
        else:
            sys.stderr.write('WARNING: No value found for %r in definition list\n'%key)
        if index:
            result += index + '\n\n'
        result += key+'\n\t'+value+'\n\n'
        for block in bullet_parts[1:]:
            lines = block.split('\n')
            if len(lines) > 1:
                for line in lines:
                    result += '\t'+line+'\n'
                result += '\n'
            else:
                if block.strip():
                    result += '\t'+block+'\n\n'
        if not index:
            result = result[:-1]                        
        return result

    def handle_code_single(self, node):
        result = ''
        for node in node.childNodes:
            if node.nodeType == node.TEXT_NODE:
                result += '\t'+node.data
            elif node.nodeName == 'span' and node.attributes['style-name'].value in ['T2']:
                result += '\t'+self.get_text(node)
            else:
                self.unexpected_node('paragraph', node)
        result = '::\n\n' + result
        return result
    
    def handle_code_line(self, node):
        result = ''
        for node in node.childNodes:
            if node.nodeType == node.TEXT_NODE:
                result += '\t'+node.data
            elif node.nodeName == 'span' and node.attributes['style-name'].value in ['T2', 'T1', 'Code_20_Inline_20_Bold']:
                result += '\t'+self.get_text(node)
            else:
                self.unexpected_node('paragraph', node)
        return result
    
    def handle_heading(self, node, underline='+'):
        result = ''
        for child in node.childNodes:
            if child.nodeType == node.TEXT_NODE:
                result += child.data
            else:
                self.unexpected_child('heading', child)
        return result+'\n'+(underline*len(result))

    def handle_row(self, node):
        node.attributes['style-name'].value = 'Text_20_body'
        result = self.handle_paragraph(node)
        cols = result.split('\t')
        return cols
        
        #for child in node.childNodes:
        #    if child.nodeName != 'tab':
        #        if isinstance(child, unicode):
        #            cols.append(child)
        #        elif child.nodeType == child.TEXT_NODE:
        #            cols.append(child.data)
        #        else:
        #return cols

    def handle_table(self, nodes):
        result = ''
        caption_node = nodes[0]
        row_nodes = nodes[2:]
        header_node = nodes[1]
        # The first node should be the caption.
        if not caption_node.attributes['style-name'].value == 'Table_20_Caption':
            sys.stderr.write('ERROR: Table started with %r node, not a caption\n'%caption_node) 
        # The last node should be a last row
        if not row_nodes[-1].attributes['style-name'].value == 'Table_20_Text_20_Last':
            sys.stderr.write("ERROR: The last table row should be 'Table_20_Text_20_Last', got %r node\n"%row_nodes[-1]) 
        row_nodes[-1].attributes['style-name'].value == 'Table_20_Text'
        # Now parse the rows
        headers = self.handle_row(header_node)
        rows = []
        for row_node in row_nodes:
            rows.append(self.handle_row(row_node))
            if not len(rows[-1]) == len(headers):
                sys.stderr.write('ERROR: One of the table rows is not the same length as the header: %r\n'%row_node)
        result += table.toRSTtable([headers]+rows, header=True, justify='left')
        result += '\n\n%s\n\n'%(self.handle_paragraph(caption_node))
        return result

    def handle_paragraph(self, node):
        index_entries = []
        result = ''
        for node in node.childNodes:
            if node.nodeType == node.TEXT_NODE:
                result += node.data
            elif node.nodeName == 'alphabetical-index-mark':
                if node.attributes.has_key('key1'):
                    index_entries.append(
                        '   single: %s; %s'%(
                            node.attributes['string-value'].value,
                            node.attributes['key1'].value,
                        )
                    )
                else:
                    index_entries.append(
                        '   single: %s'%(
                            node.attributes['string-value'].value
                        )
                    )
            elif node.nodeName == 'tab':
                result += '\t'
            elif node.nodeName == 'span':
                span = self.handle_span(node)
                if span.strip()[:2] == result.strip()[-2:] == '``' or \
                   span.strip()[:2] == result.strip()[-2:] == '**':
                    # Duplicate tags next to each other
                    result = result.strip()[:-2] + ' ' + span.strip()[2:]
                elif span.strip()[:1] == result.strip()[-1:] == '*':
                    # Duplicate tags next to each other
                    result = result.strip()[:-1] + ' ' + span.strip()[1:]
                else:
                    result += span
            else:
                self.unexpected_node('paragraph', node)
        if index_entries:
            result = '.. index :: \n'+('\n'.join(index_entries))+'\n\n' + result
        return result

    def handle_span(self, node):
        # Code_20_Inline contain URLs that shouldn't be literals and <tab />
        # elements which should end whatever is being applied and then start 
        # it again after the tab.
        if node.attributes['style-name'].value == 'Code_20_Inline' and len(node.childNodes)==1 and node.childNodes[0].nodeType == node.TEXT_NODE:
            # It is a simple literal, it could contain a URL
            content = node.childNodes[0].data
            if link(content) == content:
                return '``'+content+'``'
            else:
                # It is a link and should not be treated as a literal
                return content
        result = ''
        sub_span = False
        for child in node.childNodes:
            if child.nodeType == child.TEXT_NODE:
                result += child.data
            elif child.nodeName == 'tab':
                result += '\t'
            elif child.nodeName == 'span':
                sub_span = True
                result += handle_span(child)
            else:
                self.unexpected_child('span', child)
        if sub_span:
            sys.stderr.write('WARNING: Nested spans found, ignoring inline formatting\n')
            return result
        else:
            chars = ''
            if node.attributes['style-name'].value == 'Code_20_Inline':
                chars  = '``'
            elif node.attributes['style-name'].value == 'Code_20_Bold':
                chars  = '**'
            elif node.attributes['style-name'].value in ['T1', 'T2']:
                chars  = '*'
            else:
                self.unexpected_node('span', node)
            final = []
            result_parts = result.split('\t')
            for part in result_parts:
                if part:
                    final.append(chars+part+chars)
                else:
                    final.append(part)
            return '\t'.join(final)

    def unexpected_child(self, fn, node):
        return self._unexpected(fn, 'child', node)

    def unexpected_node(self, fn, node):
        return self._unexpected(fn, 'node', node)

    def _unexpected(self, fn, type, node):
        elem = '<%s'%node.nodeName
        if len(node.attributes):
            for k, v in node.attributes.items():
                elem += ' %s=%s'%(k, v)
        elem += '>'
        sys.stderr.write(
            (
                'handle_%s(): Unexpected %s %s'%(
                    fn,
                    type,
                    elem,
                )
            ).encode('utf-8')+'\n'
        )

#
# Function for processing a .odt file
#

def process(filename, save_stripped=False):
    # Extract the content.xml file from the .odt zip
    odt = zipfile.ZipFile(filename)
    for file in odt.infolist():
        if file.orig_filename == 'content.xml':
            content = odt.read(file.orig_filename)
    # Strip out the bits we don't want
    start_pos = content.find('</text:sequence-decls>')
    stripped = content[start_pos+22:-54]
    # Styles used only in the Front Matter which might as well just be paragraphs:
    stripped = stripped.replace('FM_20_Copyright_20_Credits_20_last', 'Text_20_body')
    stripped = stripped.replace('FM_20_Copyright_20_Credits',         'Text_20_body')
    stripped = stripped.replace('FM_20_Author',                       'Text_20_body') 
    stripped = stripped.replace('FM-Copyright_20_bold',               'T1')
    stripped = stripped.replace('FM_20_Copyright',                    'Text_20_body')
    stripped = stripped.replace('FM_20_Dedication',                   'Text_20_body') 
    # The front image isn't a full figure, just remove it
    stripped = stripped.replace('<text:p text:style-name="Production">Insert JamesGardner.tif</text:p>', '')
    # reStructuredText doesn't support soft breaks or empty paragraphs so let's remove them
    stripped = stripped.replace('<text:soft-page-break/>', '')
    stripped = stripped.replace('<text:p text:style-name="Standard"/>', '')
    # We don't need code continuation characters
    stripped = stripped.replace('<text:span text:style-name="T2">~CCC</text:span>', '')
    stripped = stripped.replace('<text:span text:style-name="T3">~CCC</text:span>', '')
    stripped = stripped.replace('<text:span text:style-name="T2">~TRA</text:span>', '-&gt;')
    stripped = stripped.replace('<text:span text:style-name="T3">~TRA</text:span>', '-&gt;')
    # We don't want comments in italics, that was a mistake by Apress
    stripped = stripped.replace('<text:p text:style-name="P3">', '<text:p text:style-name="Code_20_First">')
    stripped = stripped.replace('<text:p text:style-name="P2">', '<text:p text:style-name="Code">')
    # This is used for italicisng docstrings in code, which we won't want
    #stripped = stripped.replace('<text:p text:style-name="P4">', '<text:p text:style-name="Code">')
    # We don't want Apress's character codes
    stripped = stripped.replace('?COL?', ':')
    # Appears in Chapter 2 instead of 2 spaces.
    stripped = stripped.replace('<text:s text:c="1"/>', ' ')
    stripped = stripped.replace('<text:s text:c="2"/>', '  ')
    stripped = stripped.replace('<text:s text:c="3"/>', '   ')
    stripped = stripped.replace('<text:s text:c="4"/>', '    ')
    stripped = stripped.replace('<text:s text:c="5"/>', '     ')
    stripped = stripped.replace('<text:s text:c="6"/>', '      ')
    stripped = stripped.replace('<text:s text:c="7"/>', '       ')
    stripped = stripped.replace('<text:s text:c="8"/>', '        ')
    stripped = stripped.replace('<text:s text:c="9"/>', '         ')
    stripped = stripped.replace('<text:s text:c="10"/>', '          ')
    stripped = stripped.replace('<text:s text:c="11"/>', '           ')
    stripped = stripped.replace('<text:s text:c="12"/>', '            ')
    stripped = stripped.replace('<text:s text:c="13"/>', '             ')
    stripped = stripped.replace('<text:s text:c="14"/>', '              ')
    stripped = stripped.replace('<text:s text:c="15"/>', '               ')
    stripped = stripped.replace('<text:s text:c="16"/>', '                ')
    stripped = stripped.replace('<text:s text:c="17"/>', '                 ')
    stripped = stripped.replace('<text:s text:c="18"/>', '                  ')
    stripped = stripped.replace('<text:s text:c="19"/>', '                   ')
    stripped = stripped.replace('<text:s text:c="20"/>', '                    ')
    stripped = stripped.replace('<text:s text:c="21"/>', '                     ')
    stripped = stripped.replace('<text:s text:c="23"/>', '                       ')
    stripped = stripped.replace('<text:s text:c="24"/>', '                        ')
    stripped = stripped.replace('<text:s text:c="26"/>', '                          ')
    stripped = stripped.replace('<text:s text:c="27"/>', '                           ')
    stripped = stripped.replace('<text:s text:c="28"/>', '                            ')
    stripped = stripped.replace('<text:s text:c="31"/>', '                               ')
    stripped = stripped.replace('<text:s text:c="38"/>', '                                      ')
    stripped = stripped.replace('<text:s/>', ' ')
    stripped = stripped.replace('<text:span text:style-name="Bold_20_Text-Names"> </text:span>', ' ')
    if save_stripped:
        fp = open(filename+'.stripped.xml', 'wb')
        fp.write(stripped)
        fp.close()
    document = '<chapter>'+stripped.replace('text:', '')+'</chapter>'
    
    dom = xml.dom.minidom.parseString(document)
    
    converter = ChapterConverter()
    result = converter.handle_chapter(dom.childNodes[0])
    return result.encode('utf-8')

def main(filenames):
    done = 0
    to_process = []
    for filename in filenames:
        if not os.path.exists(filename):
            print "Skipping %r, no such file"%filename
        elif os.path.exists(filename+'.rst'):
            print "Skipping %r, the output file %r already exists"%(filename, (filename+'.rst'))
        else:
            to_process.append(filename)
    for filename in to_process:
        print "Processing %r"%filename
        result = process(filename, True) 
        fp = open(filename+'.rst', 'wb')
        fp.write(result)
        fp.close()
        print "Success, saved as %r. Done."%(filename+'.rst')
        done += 1
    if not done:
        print "ERROR: No files were converted"
    elif done == 1:
        print "Success: %d file was converted"%done 
    else:
        print "Success: %d files were converted"%done 

if __name__ == '__main__':
    if len(sys.argv) < 2:
        print "Usage: python process2.py file1.odt [file2.odt, file3.odt...] "
    else:
        main(sys.argv[1:])
        sys.exit(0)

